{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Utilities of type 'ivector' and 'imatrix'.                        }
{ Global functions:                                                          }
{    MaxRelDiam(...) : To get the maximum of the relative diameters of the   }
{                      components of an interval vector.                     }
{    UlpAcc(...)     : To check whether all components of an interval vector }
{                      have a width less than a certain number of ulps       }
{                      (ulp = unit in the last place of the mantissa).       }
{----------------------------------------------------------------------------}
module mvi_util;     { Matrix Vector Interval Utilities }
                     { -      -      -        ----      }
use
  i_ari,             { Interval arithmetic        }
  i_util;            { Utilities of type interval }

{----------------------------------------------------------------------------}
{ Compute maximum of relative diameters of the components of ivector 'v'.    }
{----------------------------------------------------------------------------}
global function MaxRelDiam ( var v: ivector ) : real;
var
  k              : integer;
  Max, RelDiam_k : real;
begin
  Max := 0;
  for k := lb(v) to ub(v) do
  begin
    RelDiam_k := RelDiam(v[k]);
    if (RelDiam_k > Max) then Max := RelDiam_k;
  end;
  MaxRelDiam := Max;
end;

{----------------------------------------------------------------------------}
{ Checks if the diameter of the interval vector 'x' is less or equal to 'n'  }
{ ulps. An ulp is an abbreviation for: units in the last place of the        }
{ mantissa.                                                                  }
{----------------------------------------------------------------------------}
global function UlpAcc ( v : ivector; n : integer ) : boolean;
var
  k : integer;
begin
  k := lb(v);
  while (k < ub(v)) and (UlpAcc(v[k],n)) do k := k+1;
  UlpAcc := UlpAcc(v[k],n);
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
