(* RCS $Id: lss.p,v 1.1.1.1 2003/11/20 08:49:04 hoeffgen Exp $ *)
(*-----------------------------------------------------------*)
(* DESCRIPTION : verified solver for linear systems          *)
(*-----------------------------------------------------------*)
MODULE lss;

use  {out,}i_ari,mv_ari,mvi_ari,lss_aprx;


CONST NULLTEST = 1E6;
      ABBRUCH  = 1E-15;
      FACTOR   = 0.31622777;
      REINS    = 1.0;
      RNULL    = 0.0;
      R1E20    = 1E20;

VAR   m,n,dim : INTEGER;


FUNCTION REL ( a,b : RVECTOR ) : REAL;
{ Berechnet den relativen Fehler von A bzgl. B. Haben A und B nicht
  dasselbe Vorzeichen oder ist B gleich 0, wird REL := 0 gesetzt 
  A ist der neue Wert einer Iteration, B der alte }
VAR i         : INTEGER;
    p,r,ai,bi : REAL;
BEGIN
   p:= RNULL;
   FOR i:= LB(a) TO UB(a) DO  { a,b muessen gleiche Indizes haben }
   BEGIN
      ai:= a[i];
      bi:= b[i];
      IF ( ai*bi <= RNULL )  OR  ( NULLTEST*ABS(ai) < ABS(bi) )
                     THEN   r:= RNULL
                     ELSE   r:= ABS( (ai-bi)/bi );
      IF r>p THEN p:= r;
   END;
   REL:= p;
END;


FUNCTION ACCUR ( VAR a : IVECTOR ) : BOOLEAN;
{ ACCUR := Genauigkeit von a allzu schlecht }
VAR i   : INTEGER;
    acc : BOOLEAN;
BEGIN
   acc:= FALSE;
   FOR i:= LB(a) TO UB(a) DO
      acc:= acc OR (a[i].INF<-R1E20) AND (a[i].SUP>R1E20);
   ACCUR := acc;
END; (* ACCUR *)


{
FUNCTION NORM( VAR V : IVECTOR ) : REAL;
{ L1-Norm von V }
{
VAR i : INTEGER;
    s : REAL;
BEGIN
   s:= RNULL;
   FOR i:= LB(V) TO UB(V) DO  s:= s +>  SUP(ABS(V[i]));
   NORM:= s;
END;


FUNCTION NORM( VAR A : IMATRIX ) : REAL;
{ Zeilensummen-Norm der Betragsmatrix von A }
{
VAR i   : INTEGER;
    n,s : REAL;
BEGIN
   n:= RNULL;
   FOR i:= LB(A,1) TO UB(A,1) DO
   BEGIN
      s:= norm(A[i]);
      IF s>n THEN n:= s;
   END;
   NORM:= n;
END;
}


PROCEDURE NULLEN_VERDACHT( VAR x1,x0 : RVECTOR );
{ x1 ist der neue, x0 der alte Wert einer Iteration. Falls sich in x1
  eine Komponente um mehr als den Faktor NULLTEST geaendert hat, so wird
  sie auf 0 gesetzt. Das gleiche gilt fuer den Fall, dass sich das Vorzeichen
  geaendert hat.          }
VAR i    : INTEGER;
    MAXX : REAL;
BEGIN                           { x0,x1 muessen gleiche Indizes haben }
  MAXX:= RNULL;
  FOR i:= LB(x1) TO UB(x1) DO
    IF ABS(x1[i])>MAXX THEN MAXX:= ABS(x1[i]);
  FOR i:= LB(x1) TO UB(x1) DO
    IF    ( (x0[i]*x1[i]<RNULL) OR (NULLTEST*ABS(x1[i])<ABS(x0[i]))  )
      AND ( MAXX>NULLTEST*ABS(x1[i]) )
    THEN  x1[i]:= RNULL;
END;


PROCEDURE LSS( VAR A : RMATRIX; VAR b : RVECTOR;  VAR Y : IVECTOR;
               VAR errcode : INTEGER;
               VAR R1,R2 : RMATRIX; VAR MB : IMATRIX; VAR FLAGS : INTEGER );
{ Das Resultat Y ist Einschliessung der Loesung von Ax=b               }
{                                                                      }
{ errcode = 0 : Y ist Einschliessung der Loesung                       }
{ errcode = 1 : keine Einschliessung erreicht, schlechte Kondition (?) }
{ errcode = 2 : keine Einschliessung moeglich, Matrix A singulaer (?)  }
{                                                                      }
{ FLAGS   = 0 : R1,R2,MB alle noch nicht vorhanden                     }
{ FLAGS   = 1 : nur R1 ist bereits vorhanden                           }
{ FLAGS   = 2 : R1 und MB hierzu sind vorhanden                        }
{ FLAGS   = 3 : R1 und R2 sind vorhanden (nicht aber MB hierzu)        }
{ FLAGS   = 4 : R1, R2 und MB hierzu sind vorhanden                    }

VAR
     D           : RMATRIX[LB(A)..UB(A),LB(A)..UB(A)];
     x0,x1,y0    : RVECTOR[LB(A)..UB(A)];
     Y1,YA,Z     : IVECTOR[LB(A)..UB(A)];
     i,j,k,err   : INTEGER;
     ready       : BOOLEAN;
     p,bound,eps : REAL;

PROCEDURE R_EINFACH;
BEGIN

{ 1. Naeherungs-Inverse berechnen }
   err:= 0;
   IF FLAGS<1 THEN
   BEGIN
      R1 := A;
      minv( R1, err );
      IF err=0 THEN FLAGS:= 1;
   END;
   IF err=0 THEN
   BEGIN

{ 2. Reelle Defektiteration. Ergebnis x1. }
      bound := 100.0*FACTOR;
      x1:= R1*b;
{
OUT( 'Reelle Defektiteration',x1 );
 }
      k:= 0;
      REPEAT
         { Iteration  x = x + R*(b-Ax)  }
         k := SUCC(k);
         x0:= x1;
         x1:= #*( b - A*x0 );
         x1:= #*( x0 + R1*x1 );
         p:= REL( x1, x0 );
         IF k>5 THEN bound := bound * FACTOR;
{
OUT( 'k = ', k );
OUT( '',x1 );
OUT( '  p   = ', p );
OUT( 'bound = ', bound );
 }
      UNTIL ( (p>=bound) AND (k>5) )  OR  (p<ABBRUCH);

      nullen_verdacht( x1,x0 );
{
 OUT( 'Naeherung', x1 );
 }
{ 3. Berechne Residuum der Naeherungsloesung
     intervallmaessig und initialisiere  Y1:= Z:= R1(b-Ax), MB:= E-R1*A  }

      y0:= #*( b - A*x1 );
      Y1:= ##( b - A*x1 - y0 );
      Y1:= ##( R1*y0 + R1*Y1 );
      Z := Y1;
{
OUT( 'Z', Z );
OUT( 'Norm des Residuums = ', NORM( Z ) );
 }
      IF Z=NULL(Z) THEN
      BEGIN
         Y      := x1;    { exakte Loesung ! ( aber nicht notw. eind. !!! ) }
         errcode:= 0;
         ready  := TRUE;
      END ELSE
      BEGIN
         IF FLAGS<2 THEN
         BEGIN
            MB   := ##( ID(A) - R1*A );
            FLAGS:= 2;
{
OUT( 'Zeilensummen-Norm von MB = ', NORM( MB ) );
 }
         END;

{ 4. Intervallmaessige Iteration bis echte Inklusion }
         k  := 0;
         eps:= 0.1;
         REPEAT
            IF k>=5 THEN eps := 5*eps;
            k := SUCC(k);
            YA:= BLOW( Y1, eps );
            Y1:= Z + MB*YA;
{
OUT( 'k = ', k );
OUT( 'YA', YA );
OUT( 'Y1', Y1 );
 }
            ready := Y1 IN YA;
         UNTIL ready OR (k>=10){ OR (YA IN Y1)} OR ACCUR(Y1);

{ 5. Ausgabe des Ergebnisses }
         IF ready THEN
         BEGIN
            Y      := x1 + Y1;
            errcode:= 0;
         END;
      END;
   END ELSE  ready:= FALSE;
   
END {=== R_EINFACH ===};


PROCEDURE R_ZWEIFACH;
BEGIN

{ 1. Naeherungs-Inverse berechnen }
   err:= 0;
   IF FLAGS<3 THEN
   BEGIN
    {  R2:= #*( R1*A - ID(A) );  
       MINV1( R2, err );       }
     
      R2:= R1*A;
      MINV( R2, err );
      IF err=0 THEN
      BEGIN
         FLAGS:= 3;
  
    {     D := #*( R1 + R2*R1 );  
          R2:= #*( R1 + R2*R1 - D );  }
     
         D := R2*R1;
         R2:= #*( R2*R1 - D );
         R1:= D;
      END;
   END;
   IF err=0 THEN
   BEGIN

{ 2. reelle Defektiteration. Ergebnis y0+x1. }
      bound := 100.0*FACTOR;
      y0:= #*( R1*b + R2*b );
      x1:= #*( R1*b + R2*b - y0 );
{
OUT( 'Reelle Defektiteration',y0 );
 }
      k:= 0;
      REPEAT
         { Iteration  x = x + (R1+R2)*(b-Ax), x = y0 + x1 }
         k := SUCC(k);
         x0:= x1;
         x1:= #*( b - A*y0 - A*x0 );
         x1:= #*( x0 + R1*x1 + R2*x1 );
         p := REL( y0+x1, y0+x0 );
         x0:= y0 + x1;
         x1:= #*( y0 + x1 - x0 );
         y0:= x0;
         IF k>5 THEN bound := bound * FACTOR;
{
OUT( 'k = ', k );
OUT( '',y0 );
OUT( '  p   = ', p );
OUT( 'bound = ', bound );
 }
      UNTIL ( (p>=bound) AND (k>5) ) OR (p<ABBRUCH);
      
{     nullen_verdacht( x1,x0 ); }
{
OUT( 'Naeherung', y0 );
 }
{ 3. Berechne Residuum der Naeherungsloesung y0 intervallmaessig
     und initialisiere Y1:= Z:= (R1+R2)(b-Ax), MB:= E-(R1+R2)*A   }

      x1:= #*( b - A*y0 );
      Y1:= ##( b - A*y0 - x1 );
      Y1:= ##( R1*x1 + R2*x1 +
               R1*Y1 + R2*Y1          );
      Z := Y1;
{
OUT( 'Z', Z );
OUT( 'Norm des Residuums = ', NORM( Z ) );
 }
      IF Z=NULL(Z) THEN
      BEGIN
         Y      := y0;    { exakte Loesung ! ( aber nicht notw. eind. !!! ) }
         errcode:= 0;
         ready  := TRUE;
      END ELSE
      BEGIN
         IF FLAGS<4 THEN
         BEGIN
            MB   := ##( ID(A) - R1*A - R2*A );
            FLAGS:= 4;
{
OUT( 'Zeilensummen-Norm von MB = ', NORM( MB ) );
 }
         END;

{ 4. Intervallmaessige Iteration bis echte Inklusion }
         k  := 0;
         eps:= 0.1;
         REPEAT
            IF k>=5 THEN eps := 5*eps;
            k := SUCC(k);
            YA:= BLOW( Y1, eps );
            Y1:= Z + MB*YA;
{
OUT( 'k = ', k );
OUT( 'YA', YA );
OUT( 'Y1', Y1 );
 }
            ready := Y1 IN YA;
         UNTIL ready OR (k>=10){ OR (YA IN Y1)} OR ACCUR(Y1);

{ 5. Ausgabe des Ergebnisses }
         IF ready THEN
         BEGIN
            Y      := y0 + Y1;
            errcode:= 0;
         END ELSE
            errcode:= 1;
      END;
   END ELSE errcode:= 2;
           
END  {=== R_ZWEIFACH ===};

BEGIN
   ready:= FALSE;

   IF FLAGS<3 THEN R_EINFACH;
   
{  Falls Misserfolg :  Neuer Versuch mit doppelt langem R = R1 + R2  }

   IF NOT ready THEN  R_ZWEIFACH;
END; {*** LSS ***}


PROCEDURE SQUARE_LSS( VAR A : RMATRIX; VAR b   : RVECTOR;
                      VAR Y : IVECTOR; VAR errcode : INTEGER );
{ Das Resultat Y ist Einschliessung der Loesung von Ax=b               }

VAR
    R1,R2 : RMATRIX[LB(A)..UB(A),LB(A)..UB(A)];
    MB    : IMATRIX[LB(A)..UB(A),LB(A)..UB(A)];
    FLAGS : INTEGER;
    RD    : REAL;
    ID    : INTERVAL;
    DP    : DOTPRECISION;

   FUNCTION DDIV( A,B,C,D : REAL ) : INTERVAL;
   { Berechnet (A*D-B*C)/(RD+ID) }
   VAR RX,RY : REAL;
       IX    : INTERVAL;
   BEGIN
      DP:= # (  A*D - B*C );
      RX:= #*( DP  );
      IX:= ##( DP  - RX );
      RY:= RX / RD;
      DDIV:= RY + ##( RX + IX - RY*RD - RY*ID )
                   / ( RD+ID );
   END; {** DDIV **}

BEGIN
   IF dim=1 THEN
   IF A[LB(A,1),LB(A,2)]=RNULL THEN errcode:= 2 ELSE
   BEGIN
      Y[LB(Y)]:= INTVAL( B[LB(B)] ) / A[LB(A,1),LB(A,2)];
      errcode:= 0;
   END ELSE
   IF dim=2 THEN
   BEGIN
      DP:= #(   A[LB(A,1),LB(A,2)]*A[UB(A,1),UB(A,2)]
              - A[LB(A,1),UB(A,2)]*A[UB(A,1),LB(A,2)] );
      IF SIGN(DP)=0 THEN errcode:= 2 ELSE
      BEGIN
         RD:= #*( DP );
         ID:= ##( DP - RD );
         IF RNULL IN (RD+ID) THEN errcode:= 1 ELSE
         BEGIN
            Y[LB(Y)]:= DDIV( B[LB(B)],A[LB(A,1),UB(A,2)],
                             B[UB(B)],A[UB(A,1),UB(A,2)] );

            Y[UB(Y)]:= DDIV( A[LB(A,1),LB(A,2)],B[LB(B)],
                             A[UB(A,1),LB(A,2)],B[UB(B)] );

            errcode:= 0;
         END;
      END;
   END ELSE
   BEGIN
      FLAGS:= 0;
      LSS( A,b,Y,errcode, R1,R2,MB,FLAGS ); 
   END;
END; {*** SQUARE_LSS ***}


PROCEDURE OVER_LSS( VAR A : RMATRIX; VAR b   : RVECTOR;
                    VAR Y : IVECTOR; VAR errcode : INTEGER );
{ Das Resultat Y ist Einschliessung der Loesung x von AH*A*x=AH*b, }
{ d.h. x ist least-square Loesung von Ax=b.                        }
{                                      | A -I  ||x| |b|            }
{ Aus Ax=b wird das (n+m)x(n+m)-System |     H || |=| | erzeugt.   } 
{                                      | 0  A  ||y| |0|            }

VAR BIG_A : RMATRIX[1..dim,1..dim];
    BIG_B : RVECTOR[1..dim];
    BIG_Y : IVECTOR[1..dim];
    i,j   : INTEGER;
BEGIN
   BIG_B:= RNULL;
   FOR i:= 1 TO m DO
   BEGIN
      BIG_A[i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO n DO BIG_A[i,j]:= A[LB(A,1)+i-1,LB(A,2)+j-1];
      BIG_A[i,n+i]:= -REINS;
   END;
   FOR i:= 1 TO n DO
   BEGIN
      BIG_A[m+i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO m DO BIG_A[m+i,n+j]:= BIG_A[j,i];
   END;
   FOR i:= 1 TO m DO  BIG_B[i]:= b[LB(b)+i-1];
   
   SQUARE_LSS(BIG_A,BIG_B,BIG_Y,errcode);
   
   FOR i:= 1 TO n DO Y[LB(Y)+i-1]:= BIG_Y[i];
END; {*** OVER_LSS ***}


PROCEDURE UNDER_LSS( VAR A : RMATRIX; VAR b   : RVECTOR;
                     VAR Y : IVECTOR; VAR errcode : INTEGER );
{ Das Resultat Y ist Einschliessung von y = AH*x mit A*AH*x=b      }
{ d.h. y ist Loesung von Ay=b mit minimaler Euklidnorm.            }
{                                      | AH -I ||x| |0|            }
{ Aus Ax=b wird das (n+m)x(n+m)-System |       || |=| | erzeugt.   } 
{                                      | 0  A  ||y| |b|            }

VAR BIG_A : RMATRIX[1..dim,1..dim];
    BIG_B : RVECTOR[1..dim];
    BIG_Y : IVECTOR[1..dim];
    i,j   : INTEGER;
BEGIN
   BIG_B:= RNULL;
   FOR i:= 1 TO n DO
   BEGIN
      BIG_A[i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO m DO BIG_A[i,j]:= A[LB(A,1)+j-1,LB(A,2)+i-1];
      BIG_A[i,m+i]:= -REINS;
   END;
   FOR i:= 1 TO m DO
   BEGIN
      BIG_A[n+i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO n DO BIG_A[n+i,m+j]:= BIG_A[j,i];
   END;
   FOR i:= 1 TO m DO  BIG_B[n+i]:= b[LB(b)+i-1];
   
   SQUARE_LSS(BIG_A,BIG_B,BIG_Y,errcode);
   
   FOR i:= 1 TO n DO Y[LB(Y)+i-1]:= BIG_Y[m+i];
END; {*** UNDER_LSS ***}


GLOBAL PROCEDURE LSS( VAR A : RMATRIX; VAR b   : RVECTOR;
                      VAR Y : IVECTOR; VAR errcode : INTEGER );
var local : integer;
BEGIN
{   TEST:= errcode = -1;
 }  errcode:= 0;

   m:= UB(A,1) - LB(A,1) + 1;
   n:= UB(A,2) - LB(A,2) + 1;

   dim:= n+m;

   IF m <> UB(b)-LB(b)+1 THEN errcode:= 3; { b : falsche Dimension }
   IF n <> UB(Y)-LB(Y)+1 THEN errcode:= 4; { Y : falsche Dimension }

   IF errcode=0 THEN
   IF m > n THEN  OVER_LSS(A,b,Y,errcode) ELSE   { ueberbestimmtes System }
   IF m < n THEN UNDER_LSS(A,b,Y,errcode)        { unterbestimmtes System }
            ELSE BEGIN
                    local:= n;
                    dim:= local;
                    SQUARE_LSS(A,b,Y,errcode);   { quadratisches System   }
                 END;
END {*** LSS ***};



PROCEDURE SQUARE_INV( VAR A : RMATRIX; VAR Y : IMATRIX;
                      VAR errcode : INTEGER );
{ Das Resultat Y ist Einschliessung der Loesung von AY=I               }
VAR
    R1,R2       : RMATRIX[LB(A)..UB(A),LB(A)..UB(A)];
    MB          : IMATRIX[LB(A)..UB(A),LB(A)..UB(A)];
    b           : RVECTOR[LB(A)..UB(A)];
    FLAGS,i,err : INTEGER;
    RD          : REAL;
    ID          : INTERVAL;
    DP          : DOTPRECISION;

   FUNCTION DDIV( A : REAL ) : INTERVAL;
   { Berechnet A/(RD+ID) }
   VAR RY : REAL;
   BEGIN
      RY:= A / RD;
      DDIV:= RY + ##( A - RY*RD - RY*ID )
                  / ( RD+ID );
   END; {** DDIV **}

BEGIN
   IF dim=1 THEN
   IF A[LB(A,1),LB(A,2)]=RNULL THEN errcode:= 2 ELSE
   BEGIN
      Y[LB(Y,1),LB(Y,2)]:= intval(1.0) / A[LB(A,1),LB(A,2)];
      errcode:= 0;
   END ELSE
   IF dim=2 THEN
   BEGIN
      DP:= #(   A[LB(A,1),LB(A,2)]*A[UB(A,1),UB(A,2)]
              - A[LB(A,1),UB(A,2)]*A[UB(A,1),LB(A,2)] );
      IF SIGN(DP)=0 THEN errcode:= 2 ELSE
      BEGIN
         RD:= #*( DP );
         ID:= ##( DP - RD );
         IF RNULL IN (RD+ID) THEN errcode:= 1 ELSE
         BEGIN
            Y[LB(Y,1),LB(Y,2)]:= DDIV(  A[UB(A,1),UB(A,2)] );
            Y[UB(Y,1),LB(Y,2)]:= DDIV( -A[UB(A,1),LB(A,2)] );
            Y[LB(Y,1),UB(Y,2)]:= DDIV( -A[LB(A,1),UB(A,2)] );
            Y[UB(Y,1),UB(Y,2)]:= DDIV(  A[LB(A,1),LB(A,2)] );

            errcode:= 0;
         END;
      END;
   END ELSE
   BEGIN
      FLAGS:= 0;
      err  := 0;
      b    := RNULL;
      FOR i:= LB(A,1) TO UB(A,1) DO
      BEGIN
         b[i]:= REINS;
         LSS( A,b,Y[LB(Y,1)+i-LB(A,1)],errcode, R1,R2,MB,FLAGS );
         if errcode>err then err:= errcode;
         b[i]:= RNULL;
      END;
      errcode:= err;
      Y:= TRANSP(Y);
   END;
END; {*** SQUARE_INV ***}


PROCEDURE OVER_INV( VAR A : RMATRIX; VAR Y : IMATRIX;
                    VAR errcode : INTEGER );
{ Das Resultat Y ist Einschliessung der Pseudoinversen A+ von A }
{                                                               }
{                   +     H  -1 H                               }
{ Fuer m>n ist Y = A  = (A A)  A   Loesung von:                 }
{                                                               }
{  | A  -I ||Y|   |I|     ( A : mxn          )                  }
{  |     H || | = | | ,   ( X = AY-I : mxm   )                  }
{  | 0  A  ||X|   |0|     ( Y = A+   : nxm   )                  }
{                         ( rechte Seite:    )                  }
{                         ( I : mxm, 0 : nxm )                  }
{                                                               }

VAR BIG_A,R1,R2   : RMATRIX[1..dim,1..dim];
    MB            : IMATRIX[1..dim,1..dim];
    BIG_B         : RVECTOR[1..dim];
    BIG_Y         : IVECTOR[1..dim];
    FLAGS,i,j,err : INTEGER;
BEGIN
   BIG_B:= RNULL;
   FOR i:= 1 TO m DO
   BEGIN
      BIG_A[i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO n DO BIG_A[i,j]:= A[LB(A,1)+i-1,LB(A,2)+j-1];
      BIG_A[i,n+i]:= -REINS;
   END;
   FOR i:= 1 TO n DO
   BEGIN
      BIG_A[m+i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO m DO BIG_A[m+i,n+j]:= BIG_A[j,i];
   END;

   FLAGS:= 0;
   err  := 0;
   FOR j:= 1 TO m DO
   BEGIN
      BIG_B[j]:= REINS;
      LSS( BIG_A,BIG_B,BIG_Y,errcode, R1,R2,MB,FLAGS );
      if errcode>err then err:= errcode;
      BIG_B[j]:= RNULL;
      FOR i:= 1 TO n DO Y[LB(Y,1)+i-1,LB(Y,2)+j-1]:= BIG_Y[i];
   END;
   errcode:= err;
END; {*** OVER_INV ***}


PROCEDURE UNDER_INV( VAR A : RMATRIX; VAR Y : IMATRIX;
                     VAR errcode : INTEGER );
{ Das Resultat Y ist Einschliessung der Pseudoinversen A+ von A }
{                                                               }
{                   +    H   H -1                               }
{ Fuer m<n ist Y = A  = A (AA )   Loesung von:                  }
{                                                               }
{     H       H                                                 }
{  | A  -I ||Y |   |I|    ( A : mxn          )                  }
{  |       || H| = | | ,  ( X = AY-I : nxn   )                  }
{  | 0  A  ||X |   |0|    ( Y = A+   : nxm   )                  }
{                         ( rechte Seite:    )                  }
{                         ( I : nxn, 0 : mxn )                  }
{                                                               }

VAR BIG_A,R1,R2   : RMATRIX[1..dim,1..dim];
    MB            : IMATRIX[1..dim,1..dim];
    BIG_B         : RVECTOR[1..dim];
    BIG_Y         : IVECTOR[1..dim];
    FLAGS,i,j,err : INTEGER;
BEGIN
   BIG_B:= RNULL;
   FOR i:= 1 TO n DO
   BEGIN
      BIG_A[i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO m DO BIG_A[i,j]:= A[LB(A,1)+j-1,LB(A,2)+i-1];
      BIG_A[i,m+i]:= -REINS;
   END;
   FOR i:= 1 TO m DO
   BEGIN
      BIG_A[n+i]:= BIG_B;  { = 0 }
      FOR j:= 1 TO n DO BIG_A[n+i,m+j]:= BIG_A[j,i];
   END;

   FLAGS:= 0;
   err  := 0;
   FOR i:= 1 TO n DO
   BEGIN
      BIG_B[i]:= REINS;
      LSS( BIG_A,BIG_B,BIG_Y,errcode, R1,R2,MB,FLAGS );
      if errcode>err then err:= errcode;
      BIG_B[i]:= RNULL;
      FOR j:= 1 TO m DO Y[LB(Y,1)+i-1,LB(Y,2)+j-1]:= BIG_Y[j];
   END;
   errcode:= err;
END; {*** UNDER_INV ***}


GLOBAL PROCEDURE INV( VAR A : RMATRIX; VAR Y : IMATRIX;
                      VAR errcode : INTEGER );
{ Das Resultat Y ist Einschliessung der Loesung von AY=I               }
var local : integer;
BEGIN
{   TEST:= errcode = -1;
 }  errcode:= 0;

   m:= UB(A,1) - LB(A,1) + 1;
   n:= UB(A,2) - LB(A,2) + 1;

   dim:= m+n;

   IF n<>UB(Y,1)-LB(Y,1)+1 THEN errcode:= 3; {Y: falsche Zeilenzahl }
   IF m<>UB(Y,2)-LB(Y,2)+1 THEN errcode:= 4; {Y: falsche Spaltenzahl}

   IF errcode=0 THEN
   IF m > n THEN  OVER_INV(A,Y,errcode) ELSE    { ueberbestimmtes System }
   IF m < n THEN UNDER_INV(A,Y,errcode)         { unterbestimmtes System }
            ELSE BEGIN
                    local:= n;
                    dim:= local;
                    SQUARE_INV(A,Y,errcode);    { quadratisches System   }
                 END;
END; {***  INV  ***}

BEGIN
END.
