/**********************************************************************/
/*                                                                    */
/*   "DP_ALLOC.C"                                                     */
/*                                                                    */
/*   Last change    == Mar-07-92 ==                                   */
/*        A.Just / R.Hammer                                           */
/*                                                                    */
/* memory allocation in handling dotprecision routines                */
/*                                                                    */
/* Feb-01-92 Just : Macro WPRINT_ON and dependend include of          */
/*                  "DP_UTIL.INC" added.                              */
/* Feb-03-92 Just : Prototype deklarations of static functions        */
/*                  removed.                                          */
/* Mar-07-92 Just : macro 'NEW_PROTO' --> 'LINT_ARGS'                 */
/*                                                                    */
/**********************************************************************/



/* #define TRACE_ON */
/* #define WPRINT_ON */




#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "impor.t"
#include        "error.h"
#include        "scan.h"
#include        "symtab.h"
#include        "idtab.h"
#include        "blktab.h"
#include        "optab.h"
#include        "bintab.h"
#include        "arrx.h"
#include        "dp_nodes.h"
#include        "dp_error.h"
#include "expor.t"
#include        "dp_alloc.h"
#include "implemen.t"



#if defined(TRACE_ON) || defined(WPRINT_ON)
#include "dp_util.inc"
#endif



/**************************************************************************/



#ifdef LINT_ARGS
tNODE *d_getnode(void)
#else
tNODE *d_getnode()
#endif

/*--------------------------------------------------------------------*/
/* This function allocates the memory for a node of the working tree. */
/* The node is added to the tree as a child of the node 'father' is   */
/* pointing to. The father node may not have a child tree on the side */
/* the New node will be added.                                        */
/* All values will be set to NULL.                                    */
/* The return value is a pointer to the New node.                     */
/*--------------------------------------------------------------------*/

{
     register tNODE *ret;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_getnode()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* memory allocation                                             */
     /* on failure call routine for critical errors                   */
     /*---------------------------------------------------------------*/

     if( (ret=(tNODE*)malloc(sizeof(tNODE))) == NULL )
          d_critical("Memory allocation failed when creating work tree!");



     /*---------------------------------------------------------------*/
     /* initialisation of the node with NULL-elements                 */
     /*---------------------------------------------------------------*/

     ret->left           = NULL;
     ret->right          = NULL;
     ret->father         = NULL;
     ret->pos.col        = 0;
     ret->pos.line       = 0;
     ret->structure      = no_structype;
     ret->basetype       = no_arithtype;
     ret->kind           = no_nodetype;
     ret->namekind       = no_namekind;
     ret->name.direct    = no_nodename;



     /*---------------------------------------------------------------*/
     /* return pointer to the New node                                */
     /*---------------------------------------------------------------*/

     return(ret);

} /* d_getnode() */



/**************************************************************************/



#ifdef LINT_ARGS
char *d_getstring(char *in_string)
#else
char *d_getstring(in_string)
char *in_string;
#endif

/*--------------------------------------------------------------------*/
/* This function allocates memory for a string and makes a copy of    */
/* 'in_string'. It returns a pointer to the copy.                     */
/*--------------------------------------------------------------------*/
/* in_string = pointer to the string                                  */
/*--------------------------------------------------------------------*/

{
     char *out_string;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_getstring()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* memory allocation                                             */
     /* on failure call routine for critical errors                   */
     /*---------------------------------------------------------------*/

     if( (out_string=(char*)malloc(strlen(in_string)+1)) == (char*)0 )
          d_critical("Memory allocation failed when creating work tree!");



     /*---------------------------------------------------------------*/
     /* copy 'in_string' to 'out_string' and return pointer           */
     /*---------------------------------------------------------------*/

     (void) strcpy(out_string,in_string);
     return(out_string);

} /* d_getstring() */



/**************************************************************************/



#ifdef LINT_ARGS
static void d_freenode(tNODE *ptr)
#else
static void d_freenode(ptr)
register tNODE *ptr;
#endif

/*--------------------------------------------------------------------*/
/* free memory allocated by the node                                  */
/*--------------------------------------------------------------------*/
/* ptr = pointer to the work node                                     */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_freenode()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( ptr == NULL ){d_checkerr(cISNULL,pos,"ptr");return;}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* return memory to operating system                             */
     /*---------------------------------------------------------------*/

     if( ptr->namekind == d_charptr )
          free( ptr->name.str );

     if( ptr->namekind == d_varblk )
          free(ptr->name.varinfo);

     free(ptr);

} /* d_freenode() */



/**************************************************************************/



#ifdef LINT_ARGS
tNODE *d_instnode(tNODE *ptr,char fdir,char cdir)
#else
tNODE *d_instnode(ptr,fdir,cdir)
register tNODE *ptr;
register char  fdir,cdir;
#endif

/*--------------------------------------------------------------------*/
/* This function inserts a node into the tree, either as right or as  */
/* left child of the father node.                                     */
/* The New node is initialised with the given arguments, all pointers */
/* are set by the routine.                                            */
/* 'fdir' describes the addition of the New node in relation to the   */
/* father node, 'cdir' the addition of the child nodes in relation to */
/* the New node ( for further see below).                             */
/* The return value is a pointer to the New node.                     */
/*--------------------------------------------------------------------*/
/* ptr = pointer to father node                                       */
/* fdir, cdir = 'l'->add as left, 'r'->add as rigth child             */
/*--------------------------------------------------------------------*/

{
     register tNODE *New,*child;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_instnode()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( ptr == NULL )
          d_critical("Pointer argument is NULL in 'd_instnode()'!");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* get New work node                                             */
     /*---------------------------------------------------------------*/

     New   = d_getnode();

     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( New == NULL )
          d_critical("Allocation returned NULL-pointer in 'd_instnode()'!");
#endif
     /*---------------------------------------------------------------*/

     

     New->father = ptr;



     /*---------------------------------------------------------------*/
     /* insert New node as left or right child                        */
     /*---------------------------------------------------------------*/

     if( fdir == 'l' ){
          child = ptr->left;
          /*----------------------------------------------------------*/
#ifdef CHECK
          if( ptr->left == NULL )
               d_critical("Left child is NULL-Pointer in 'd_instnode()'!");
#endif
          /*----------------------------------------------------------*/
          ptr->left->father = New;
          ptr->left = New;}
     else{
          child = ptr->right;
          /*----------------------------------------------------------*/
#ifdef CHECK
          if( ptr->right == NULL )
               d_critical("Right child is NULL-Pointer in 'd_instnode()'!");
#endif
          /*----------------------------------------------------------*/
          ptr->right->father = New;
          ptr->right = New;}



     /*---------------------------------------------------------------*/
     /* add old child as left or right child of the New node          */
     /*---------------------------------------------------------------*/

     if( cdir == 'l' )
          New->left = child;
     else
          New->right = child;



     /*---------------------------------------------------------------*/
     /* return pointer to New node                                    */
     /*---------------------------------------------------------------*/

     return(New);

} /* d_instnode() */
          


/**************************************************************************/



#ifdef LINT_ARGS
tNODE *d_extrnode(tNODE *ptr)
#else
tNODE *d_extrnode(ptr)
register tNODE *ptr;
#endif

/*--------------------------------------------------------------------*/
/* This function extracts a node out of the working tree. The node    */
/* MUST have exactly ONE CHILD, thous means that either the pointer   */
/* to the left OR to the right child must be set to NULL.             */
/* Also the node MUST have a fathernode.                              */
/* If the function is forced to work on a node that doesn't fullfill  */
/* this, 'd_critical()' will be called and no work will be done on    */
/* return of the function.                                            */
/* The subtree of the extracted node will be added in the same posi-  */
/* tion the node had.                                                 */
/* The return value is a pointer to the child of the node to be ex-   */
/* tracted.                                                           */
/*--------------------------------------------------------------------*/
/* ptr = pointer to the work node                                     */
/*--------------------------------------------------------------------*/

{
     register tNODE *child;



     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_extrnode()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if( ptr == NULL )
          d_critical("Pointer argument is NULL in 'd_extrnode()'!");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
#ifdef CHECK
     if(  ( (ptr->left!=NULL) && (ptr->right!=NULL) ) ||
          ( (ptr->left==NULL) && (ptr->right==NULL) ) ||
          (       ptr->father == NULL               )    ){
          d_checkerr(cILNODE,ptr->pos,NULL);
          if( ptr->left  != NULL ) return(ptr->left);
          if( ptr->right != NULL ) return(ptr->right);
          return(ptr);}
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* get child of the node                                         */
     /*---------------------------------------------------------------*/

     if( ptr->left == NULL )
          child = ptr->right;
     else
          child = ptr->left;



     /*---------------------------------------------------------------*/
     /* get direction in relation to fathernode                       */
     /*---------------------------------------------------------------*/

     if( ptr == ptr->father->left ){
          ptr->father->left = child;
          child->father = ptr->father;}
     else{
          ptr->father->right = child;
          child->father = ptr->father;} /* endif */



     /*---------------------------------------------------------------*/
     /* deallocate work node and return pointer to its child          */
     /*---------------------------------------------------------------*/

     d_freenode(ptr);
     return(child);

} /* d_extrnode() */



/**************************************************************************/



#ifdef LINT_ARGS
void d_freetree(tNODE *ptr)
#else
void d_freetree(ptr)
register tNODE *ptr;
#endif

/*--------------------------------------------------------------------*/
/* This function disallocates the work tree pointed at.               */
/*--------------------------------------------------------------------*/
/* ptr = pointer to the tree (or subtree) to be freed                 */
/*--------------------------------------------------------------------*/

{
     /*---------------------------------------------------------------*/
#ifdef TRACE_ON
          d_trace("'d_freetree()'");
#endif
     /*---------------------------------------------------------------*/



     /*---------------------------------------------------------------*/
     /* a NULL-pointer means the end of the binary search             */
     /*---------------------------------------------------------------*/

     if( ptr == NULL )return;



     /*---------------------------------------------------------------*/
     /* release left and right subtree if exist                       */
     /*---------------------------------------------------------------*/

     if( ptr->left  != NULL )
     d_freetree(ptr->left);

     if( ptr->right != NULL )
     d_freetree(ptr->right);



     /*---------------------------------------------------------------*/
     /* release the work node itself                                  */
     /*---------------------------------------------------------------*/

     d_freenode(ptr);

} /* d_freetree() */



/**************************************************************************/
